////////////////////////////////////////////////////////////////////////////////
// Copyright (c) 2017 Josef Spidlen, Ph.D., FlowJo, LLC
//
// License
// The software is distributed under the terms of the 
// Artistic License 2.0
// http://www.r-project.org/Licenses/Artistic-2.0
// 
// Disclaimer
// This software and documentation come with no warranties of any kind.
// This software is provided "as is" and any express or implied 
// warranties, including, but not limited to, the implied warranties of
// merchantability and fitness for a particular purpose are disclaimed.
// In no event shall the  copyright holder be liable for any direct, 
// indirect, incidental, special, exemplary, or consequential damages
// (including but not limited to, procurement of substitute goods or 
// services; loss of use, data or profits; or business interruption)
// however caused and on any theory of liability, whether in contract,
// strict liability, or tort arising in any way out of the use of this 
// software.    
//////////////////////////////////////////////////////////////////////////////

package com.flowjo.plugin.FlowSOM;
import java.awt.Component;
import java.awt.Dimension;
import java.awt.event.ItemEvent;
import java.awt.event.ItemListener;
import java.io.File;
import java.io.FileNotFoundException;
import java.io.FilenameFilter;
import java.net.URL;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Scanner;

import javax.swing.Box;
import javax.swing.DefaultListModel;
import javax.swing.Icon;
import javax.swing.ImageIcon;
import javax.swing.JCheckBox;
import javax.swing.JOptionPane;
import javax.swing.JScrollPane;
import javax.swing.event.ListSelectionEvent;
import javax.swing.event.ListSelectionListener;

import com.flowjo.plugin.FlowSOM.utils.DoubleNumberFormatter;
import com.treestar.flowjo.application.workspace.Workspace;
import com.treestar.flowjo.application.workspace.manager.FJApplication;
import com.treestar.flowjo.application.workspace.manager.WSDocument;
import com.treestar.flowjo.core.Sample;
import com.treestar.flowjo.engine.utility.R_Algorithm;
import com.treestar.lib.FJPluginHelper;
import com.treestar.lib.core.ExportFileTypes;
import com.treestar.lib.core.ExternalAlgorithmResults;
import com.treestar.lib.data.StringUtil;
import com.treestar.lib.gui.FJList;
import com.treestar.lib.gui.GuiFactory;
import com.treestar.lib.gui.HBox;
import com.treestar.lib.gui.numberfields.RangedDoubleTextField;
import com.treestar.lib.gui.numberfields.RangedIntegerTextField;
import com.treestar.lib.gui.panels.FJLabel;
import com.treestar.lib.gui.swing.FJCheckBox;
import com.treestar.lib.gui.swing.FJComboBox;
import com.treestar.lib.xml.SElement;

public class FlowSOM extends R_Algorithm {

	private static final String pluginVersion = "1.4";
	private static final String pluginName = "FlowSOM";
    public static final String One = "1";
    public static final String Zero = "0";
    public static final String True = "true";
    public static final String False = "false";
    public static final String cellIdParName = "CellId";

	private RangedIntegerTextField fNumClustField = null, fDimXField = null, fDimYField = null;
	private RangedDoubleTextField fNodeScaleField = null;
    private FJComboBox fViewCombo = new FJComboBox(viewComboOptionsHuman);
    private FJComboBox fPlotChannelsCombo = null;
    private FJComboBox fApplyOnPrevCombo = null;
    private FJCheckBox fMetaclustersBackgroundCheckbox = null, fScaleOptionCheckbox = null, 
    		fCreatePdfOptionCheckbox = null, fFixedNodeSizeOptionCheckbox = null, fAddMFIsTableCheckbox = null;

    // Note that the order of items in viewComboOptionsHuman must correspond to viewComboOptions
    public static final String[] viewComboOptionsHuman = new String[] { "Minimum Spanning Tree", "t-SNE", "Grid", "All 3 options" };
    public static final String[] viewComboOptions = new String[] { "MST", "tSNE", "grid", "all" };
    
    private static final int fixedLabelWidth  = 130;
    private static final int fixedFieldWidth  = 75;
    private static final int fixedLabelHeigth = 25;
    private static final int fixedFieldHeigth = 25;
    private static final int fixedToolTipWidth = 300;
    private static final int fixedComboWidth   = 250;
    private static final int hSpaceHeigth = 5;
    private SElement fsElement = null;
    
    private static final String applyOnPrevLabel      = "Apply on map";
    private static final String applyOnPrevTooltip    = "If you have executed FlowSOM before, you can apply new data to a map generated by previous runs of FlowSOM. A new FlowSOM object will be created with the same grid but new mapping, node sizes and mean values.";
    private static final String keepResultsTooltip    = "Keep a file that shows execution of the script.";
    private static final String keepResultsLabel      = "Save the R script and output messages";
    private static final String numClustLabel         = "Number of meta clusters";
    private static final String numClustLabelTooltip  = "The number of meta clusters is the number of groups of cells to look for.";
    private static final String dimXLabel             = "SOM grid size (W x H)";
    private static final String dimXTooltip           = "Width of the grid for building the self-organizing map.";
    private static final String dimYTooltip           = "Hight of the grid for building the self-organizing map.";
    private static final String fixedNodeSizeLabel    = "Fixed node size";
    private static final String fixedNodeSizeTooltip  = "Should each node be drawn using the same size? Alternativelly, the size of the node will depend on the number of cells it represents.";
    private static final String addMFIsTableLabel     = "Include MFIs table";
    private static final String addMFIsTableTooltip   = "Should a table of scaled MFI values be included in the result when displayed in the layout editor (a CSV file with MFIs will be produced regardless).";
    
    private static final String nodeScaleLabel        = "Node scale";
    private static final String nodeScaleTooltip      = "Node scaling factor, 100% by default, smaller values will decrease the node sizes, large values will increase the node sizes.";
    
    private static final String displayLabel          = "Display options";
    private static final String orPerformFlowSOMLabel = "or perform new FlowSOM";
    private static final String viewLabel             = "View layout";
    private static final String viewTooltip           = "Select the layout for clusters in the resulting figure.";
    private static final String backgroundLabel       = "Background metaclusters";
    private static final String backgroundTooltip     = "Use the metacluster number for background coloring?";
    private static final String scaleLabel            = "Scale";
    private static final String scaleTooltip          = "Should the data be scaled by the FlowSOM function?";
    private static final String createPdfLabel        = "Create PDF";
    private static final String createPdfTooltip      = "Should a PDF output also be created in the plugin working directory?";
    private static final String plotLabel             = "Plot channels";
    private static final String plotTooltip           = "Select how to color nodes in the resulting figure. This channel needs to be among the parameters selected above or else we will just plot all available parameters as piecharts.";
    
    public static final String allChannelsText       = "All as piecharts"; 
    
    public static final String numClustOptionName       = "nClus";
    public static final String viewOptionName           = "view";
    public static final String bgMetaClustersOptionName = "bgMetaClusters";
    public static final String plotOptionName           = "plot";
    public static final String scaleOptionName          = "scale";
    public static final String createPdfOptionName      = "pdf";
    public static final String addMFIsTableOptionName   = "MFIsInTable";
    public static final String xDimOptionName           = "xdim";
    public static final String yDimOptionName           = "ydim";
    public static final String fixedNodeSizeOptionName  = "fixedNodeSize";
    public static final String nodeScaleOptionName      = "nodeScale";
    public static final String applyOnPrevOptionName    = "applyOn"; // "None" or file path to an RData file with a FlowSOM object
    public static final String pluginFolderAttName      = "pluginFolder";
    
    public static final String RDataFileExtension       = ".RData";
    public static final String RDataFileSuffix          = ".csv.flowSOM.csv.RData";
    public static final String CSVwithParsFileSuffix    = ".csv.flowSOM.csv.pars.csv";

    public static final int defaultNumClust = 8; // Min xDim and yDim is 3, I want the default of num clust to be no more than 3*3 so that I don't have to do additional checks in FlowSOMRFlowCalc
    public static final int defaultXDim = 10;
    public static final int defaultYDim = 10;   
    public static final double defaultNodeScale = 100.0; // Node scaling 100% by default
    public static final boolean defaultMetaclustersBackground = true;
    public static final boolean dafaultScale = false;
    public static final boolean dafaultCreatePDF = true;
    public static final boolean defaultFixedNodeSize = false;
    public static final String defaultApplyOnPrev = "None";
    public static final boolean defaultAddMFIsToTable = false;    
    
    private boolean fMetaclustersBackground = defaultMetaclustersBackground;
    private boolean fAddMFIsToTable = defaultAddMFIsToTable;
    private boolean fScale = dafaultScale;
    private boolean fCreatePdf = dafaultCreatePDF;
    private boolean fFixedNodeSize = defaultFixedNodeSize;
    private int fnclus = defaultNumClust, fndimx = defaultXDim, fndimy = defaultYDim;
    private double fNodeScale = defaultNodeScale;
    
    private static final String channelsLabelLine1 = "FCS channels to be used by flowSOM. Select multiple items by pressing the Shift";
    private static final String channelsLabelLine2 = "key or toggle items by holding the Ctrl (or Cmd) keys.";
    
    private static final String citingLabelLine1   = "If you are using FlowSOM, please cite the authors of the algorithm:";
    private static final String citingLabelLine2   = "Van Gassen S, Callebaut B, Van Helden MJ, Lambrecht BN, Demeester P,";
    private static final String citingLabelLine3   = "Dhaene T, Saeys Y. FlowSOM: Using self-organizing maps for visualization";
    private static final String citingLabelLine4   = "and interpretation of cytometry data. Cytometry A. 2015; 87(7):636-45.";
    
    protected static final String sIconName = "images/FlowSOMIcon.png";
    private static Icon myIcon = null;
	private static final String Failed = "Failed";
	
    public FlowSOM() 
	{
		super(pluginName);
		fShowOutput = true;
	}

	public String getName() 
	{
		return(pluginName);
	}

	public String getVersion() 
	{
		return(pluginVersion);
	}
	
	@Override
	public Icon getIcon()
	{
		if (myIcon == null)
		{
			URL url = getClass().getClassLoader().getResource(getIconName());
			if (url != null)
				myIcon = new ImageIcon(url);
		}
		return myIcon;
	}
	
	@Override
	protected String getIconName()
	{
		return sIconName;
	}
	
	@Override
	protected boolean showClusterInputField()
	{
		return false;
	}
	
	@Override
	public ExportFileTypes useExportFileType() {
		return ExportFileTypes.CSV_CHANNEL;
	}

	public ExternalAlgorithmResults invokeAlgorithm(SElement fcmlQueryElement, File sampleFile, File outputFolder) 
	{
        ExternalAlgorithmResults results = new ExternalAlgorithmResults();
                
        if(!sampleFile.exists()) 
        {
            // results.setErrorMessage("Input file did not exist"); // We purposely don't want to set the error as there may be undesirable side-effects
        	JOptionPane.showMessageDialog(null, "Input file did not exist", "FlowSOM error", JOptionPane.ERROR_MESSAGE);
            results.setWorkspaceString(FlowSOM.Failed);
            return results;
        } 
        else
        {
        	// Let's force recalculation all the time because it's relatively quick and we don't seem to handle
        	// checkUseExistingFiles well (i.e., if input settings change a bit, we still tend to return previous
        	// results instead of recalculating
        	// checkUseExistingFiles(fcmlQueryElement);
        	fUseExistingFiles = false;
        	
            String sampleName = StringUtil.rtrim(sampleFile.getName(), ".fcs");
            
            FlowSOMRFlowCalc calculator = new FlowSOMRFlowCalc();
            File flowSOMResult = calculator.runFlowSOM(sampleFile, sampleName, preprocessCompParameterNames(), fOptions, outputFolder.getAbsolutePath(), useExistingFiles());
            calculator.deleteScriptFile();
            checkROutFile(calculator);

            results.setCSVFile(flowSOMResult);
            
            /*
            // TODO: This could be used instead of results.setCSVFile(flowSOMResult) but would need some adjustments to make it work on
            // subpopulations where the derived parameter name should reflect the gating hierarchy
            if (flowSOMResult != null && flowSOMResult.exists()) {
            	PluginHelper.createClusterParameter(results, foo, flowSOMResult);
            	addGatingML(results);
            } else {
            	results.setErrorMessage("flowSOM has failed to produce the expected derived parameter.");
            }*/
            
            long timeStamp = System.currentTimeMillis();
            // Timestamp the image file
            try {
            	File imageFile = new File(flowSOMResult.getAbsolutePath().toString() + ".png");
            	File timeStampledImageFile = new File(flowSOMResult.getAbsolutePath().toString() + "." + timeStamp + ".png");
            	if (imageFile.exists() && !timeStampledImageFile.exists()) {
            		if (imageFile.renameTo(timeStampledImageFile))
            			results.setImageURL(timeStampledImageFile.toURI().toURL());
            		else {
            			// results.setErrorMessage("FlowSOM failed to timestamp the image result."); // We purposely don't want to set the error as there may be undesirable side-effects
                    	JOptionPane.showMessageDialog(null, "FlowSOM failed to timestamp the image result.", "FlowSOM error", JOptionPane.ERROR_MESSAGE);
            		}
            	}
            }
			catch (Exception e) { }
            try {
            	File imageFile = new File(flowSOMResult.getAbsolutePath().toString() + ".PopLabels.png");
            	File timeStampledImageFile = new File(flowSOMResult.getAbsolutePath().toString() + "." + timeStamp + ".PopLabels.png");
            	if (imageFile.exists() && !timeStampledImageFile.exists())
            		imageFile.renameTo(timeStampledImageFile);
            }
			catch (Exception e) { }
            try {
            	File imageFile = new File(flowSOMResult.getAbsolutePath().toString() + ".ClLabels.png");
            	File timeStampledImageFile = new File(flowSOMResult.getAbsolutePath().toString() + "." + timeStamp + ".ClLabels.png");
            	if (imageFile.exists() && !timeStampledImageFile.exists())
            		imageFile.renameTo(timeStampledImageFile);
            }
			catch (Exception e) { }
            try {
            	File imageFile = new File(flowSOMResult.getAbsolutePath().toString() + "_tSNE.png");
            	File timeStampledImageFile = new File(flowSOMResult.getAbsolutePath().toString() + "." + timeStamp + ".tSNE.png");
            	if (imageFile.exists() && !timeStampledImageFile.exists())
            		imageFile.renameTo(timeStampledImageFile);
            }
			catch (Exception e) { }
            try {
            	File imageFile = new File(flowSOMResult.getAbsolutePath().toString() + "_tSNE.PopLabels.png");
            	File timeStampledImageFile = new File(flowSOMResult.getAbsolutePath().toString() + "." + timeStamp + ".tSNE.PopLabels.png");
            	if (imageFile.exists() && !timeStampledImageFile.exists())
            		imageFile.renameTo(timeStampledImageFile);
            }
			catch (Exception e) { }
            try {
            	File imageFile = new File(flowSOMResult.getAbsolutePath().toString() + "_tSNE.ClLabels.png");
            	File timeStampledImageFile = new File(flowSOMResult.getAbsolutePath().toString() + "." + timeStamp + ".tSNE.ClLabels.png");
            	if (imageFile.exists() && !timeStampledImageFile.exists())
            		imageFile.renameTo(timeStampledImageFile);
            }
			catch (Exception e) { }
            try {
            	File imageFile = new File(flowSOMResult.getAbsolutePath().toString() + "_grid.png");
            	File timeStampledImageFile = new File(flowSOMResult.getAbsolutePath().toString() + "." + timeStamp + ".grid.png");
            	if (imageFile.exists() && !timeStampledImageFile.exists())
            		imageFile.renameTo(timeStampledImageFile);

			}
			catch (Exception e) { }
            try {
            	File imageFile = new File(flowSOMResult.getAbsolutePath().toString() + "_grid.PopLabels.png");
            	File timeStampledImageFile = new File(flowSOMResult.getAbsolutePath().toString() + "." + timeStamp + ".grid.PopLabels.png");
            	if (imageFile.exists() && !timeStampledImageFile.exists())
            		imageFile.renameTo(timeStampledImageFile);

			}
			catch (Exception e) { }
            try {
            	File imageFile = new File(flowSOMResult.getAbsolutePath().toString() + "_grid.ClLabels.png");
            	File timeStampledImageFile = new File(flowSOMResult.getAbsolutePath().toString() + "." + timeStamp + ".grid.ClLabels.png");
            	if (imageFile.exists() && !timeStampledImageFile.exists())
            		imageFile.renameTo(timeStampledImageFile);

			}
			catch (Exception e) { }
            
            // Timestamp the PDF and other files should those exist
            try {
            	File pdfFile = new File(flowSOMResult.getAbsolutePath().toString() + ".pdf");
            	File timeStampledPdfFile = new File(flowSOMResult.getAbsolutePath().toString() + "." + timeStamp + ".pdf");
            	if (pdfFile.exists() && !timeStampledPdfFile.exists()) 
            		pdfFile.renameTo(timeStampledPdfFile);
			}
			catch (Exception e) { }
            try {
            	File pdfFile = new File(flowSOMResult.getAbsolutePath().toString() + ".PopLabels.pdf");
            	File timeStampledPdfFile = new File(flowSOMResult.getAbsolutePath().toString() + "." + timeStamp + ".PopLabels.pdf");
            	if (pdfFile.exists() && !timeStampledPdfFile.exists())
            		pdfFile.renameTo(timeStampledPdfFile);
            }
			catch (Exception e) { }
            try {
            	File pdfFile = new File(flowSOMResult.getAbsolutePath().toString() + ".ClLabels.pdf");
            	File timeStampledPdfFile = new File(flowSOMResult.getAbsolutePath().toString() + "." + timeStamp + ".ClLabels.pdf");
            	if (pdfFile.exists() && !timeStampledPdfFile.exists())
            		pdfFile.renameTo(timeStampledPdfFile);
            }
			catch (Exception e) { }
            try {
            	File pdfFile = new File(flowSOMResult.getAbsolutePath().toString() + "_tSNE.pdf");
            	File timeStampledPdfFile = new File(flowSOMResult.getAbsolutePath().toString() + "." + timeStamp + ".tSNE.pdf");
            	if (pdfFile.exists() && !timeStampledPdfFile.exists()) 
            		pdfFile.renameTo(timeStampledPdfFile);
			}
			catch (Exception e) { }
            try {
            	File pdfFile = new File(flowSOMResult.getAbsolutePath().toString() + "_tSNE.PopLabels.pdf");
            	File timeStampledPdfFile = new File(flowSOMResult.getAbsolutePath().toString() + "." + timeStamp + ".tSNE.PopLabels.pdf");
            	if (pdfFile.exists() && !timeStampledPdfFile.exists()) 
            		pdfFile.renameTo(timeStampledPdfFile);
			}
			catch (Exception e) { }
            try {
            	File pdfFile = new File(flowSOMResult.getAbsolutePath().toString() + "_tSNE.ClLabels.pdf");
            	File timeStampledPdfFile = new File(flowSOMResult.getAbsolutePath().toString() + "." + timeStamp + ".tSNE.ClLabels.pdf");
            	if (pdfFile.exists() && !timeStampledPdfFile.exists()) 
            		pdfFile.renameTo(timeStampledPdfFile);
			}
			catch (Exception e) { }
            try {
            	File pdfFile = new File(flowSOMResult.getAbsolutePath().toString() + "_grid.pdf");
            	File timeStampledPdfFile = new File(flowSOMResult.getAbsolutePath().toString() + "." + timeStamp + ".grid.pdf");
            	if (pdfFile.exists() && !timeStampledPdfFile.exists()) 
            		pdfFile.renameTo(timeStampledPdfFile);
			}
			catch (Exception e) { }
            try {
            	File pdfFile = new File(flowSOMResult.getAbsolutePath().toString() + "_grid.PopLabels.pdf");
            	File timeStampledPdfFile = new File(flowSOMResult.getAbsolutePath().toString() + "." + timeStamp + ".grid.PopLabels.pdf");
            	if (pdfFile.exists() && !timeStampledPdfFile.exists()) 
            		pdfFile.renameTo(timeStampledPdfFile);
			}
			catch (Exception e) { }            
            try {
            	File pdfFile = new File(flowSOMResult.getAbsolutePath().toString() + "_grid.ClLabels.pdf");
            	File timeStampledPdfFile = new File(flowSOMResult.getAbsolutePath().toString() + "." + timeStamp + ".grid.ClLabels.pdf");
            	if (pdfFile.exists() && !timeStampledPdfFile.exists()) 
            		pdfFile.renameTo(timeStampledPdfFile);
			}
			catch (Exception e) { }            
            try {
            	File pdfFile = new File(flowSOMResult.getAbsolutePath().toString() + ".PopHm.pdf");
            	File timeStampledPdfFile = new File(flowSOMResult.getAbsolutePath().toString() + "." + timeStamp + ".PopHm.pdf");
            	if (pdfFile.exists() && !timeStampledPdfFile.exists()) 
            		pdfFile.renameTo(timeStampledPdfFile);
			}
			catch (Exception e) { }
            try {
            	File pdfFile = new File(flowSOMResult.getAbsolutePath().toString() + ".ClHm.pdf");
            	File timeStampledPdfFile = new File(flowSOMResult.getAbsolutePath().toString() + "." + timeStamp + ".ClHm.pdf");
            	if (pdfFile.exists() && !timeStampledPdfFile.exists()) 
            		pdfFile.renameTo(timeStampledPdfFile);
			}
			catch (Exception e) { }
            try {
            	File csvFile = new File(flowSOMResult.getAbsolutePath().toString() + ".PopMFI.csv");
            	File timeStampledCsvFile = new File(flowSOMResult.getAbsolutePath().toString() + "." + timeStamp + ".PopMFI.csv");
            	if (csvFile.exists() && !timeStampledCsvFile.exists()) 
            		csvFile.renameTo(timeStampledCsvFile);
			}
			catch (Exception e) { }
            try {
            	File csvFile = new File(flowSOMResult.getAbsolutePath().toString() + ".ClMFI.csv");
            	File timeStampledCsvFile = new File(flowSOMResult.getAbsolutePath().toString() + "." + timeStamp + ".ClMFI.csv");
            	if (csvFile.exists() && !timeStampledCsvFile.exists()) 
            		csvFile.renameTo(timeStampledCsvFile);
			}
			catch (Exception e) { }
            // End of timestamping output
            
            String sAddMFIsToTable = fOptions.get(FlowSOM.addMFIsTableOptionName);
            fAddMFIsToTable = true;
            if (sAddMFIsToTable == null || sAddMFIsToTable.isEmpty() || FlowSOM.Zero.equals(sAddMFIsToTable) || FlowSOM.False.equals(sAddMFIsToTable))
            	fAddMFIsToTable = false; // f is the default
            
            if (fAddMFIsToTable) try {
            	File csvFile = new File(flowSOMResult.getAbsolutePath().toString() + "." + timeStamp + ".PopMFI.csv");
            	if (csvFile.exists()) {
            		addTableFromCSVToResults(results, csvFile);
            	}
			}
			catch (Exception e) { }
                                    
            return results;
        }
	}
	
	/*
	private void addGatingML(ExternalAlgorithmResults results)
	{
		// Gate on the "flowSOM" parameter
		String sNumClus = fOptions.get(FlowSOM.numClustOptionName);
		int numClus = FlowSOM.defaultNumClust;
        try {
        	if ((Integer.parseInt(sNumClus) < 0) || (Integer.parseInt(sNumClus) > 1000))
        		numClus = Integer.parseInt(sNumClus);
        } catch (Exception e) {
        	numClus = FlowSOM.defaultNumClust;
        }
        
		SElement gate = new SElement("gating:Gating-ML");
		for (int i = 1; i <= numClus; i++) {
			SElement myFlowSOMGate = new SElement("gating:RectangleGate");
			myFlowSOMGate.setString("gating:id", "FlowSOM_" + i);
			gate.addContent(myFlowSOMGate);
			SElement dimElem1 = new SElement("gating:dimension");
			dimElem1.setDouble("gating:min", i * 100 - 10);
			dimElem1.setDouble("gating:max", i * 100 + 10);
			myFlowSOMGate.addContent(dimElem1);
			SElement fcsDimElem1 = new SElement("data-type:fcs-dimension");
			fcsDimElem1.setString("data-type:name", derParName);
			dimElem1.addContent(fcsDimElem1);			
		}

		results.setGatingML(gate.toString());
	}
	*/
	
	private void addTableFromCSVToResults(ExternalAlgorithmResults results, File csvFile)
	{
    	if (csvFile.exists()) {
    	    Scanner scan;
			try {
				int nCols = 0;
				scan = new Scanner(csvFile);
				if(scan.hasNextLine()) {
					String line = scan.nextLine();
					String[] values = line.split(",");
					if (values != null) {
						nCols = values.length;
						for (int i = 0; i < nCols; i++)
							values[i] = values[i].replaceAll("^\"|\"$", "").replaceAll("<", "(").replaceAll(">", ")");
						results.setTableHeaders(values);					
					}
				}
				ArrayList<double[]> listOfLines = new ArrayList<double[]>();
        	    if (nCols > 0) {
        	    	while (scan.hasNextLine()) {
        	    		String line = scan.nextLine();
        	    		String[] values = line.split(",");
        	    		if (values != null && values.length == nCols) { // Only process lines with the correct number of columns
        	    			double dV[] = new double[nCols];
        	    			for (int i = 0; i < values.length; i++) {  
        	    				dV[i] = Double.parseDouble(values[i].replaceAll("^\"|\"$", ""));
        	    			}
        	    			listOfLines.add(dV);
        	    		}
        	    	}
        	    	double[][] v = new double[listOfLines.size()][nCols];
        	    	for (int i = 0; i < listOfLines.size(); i++) {
        	    		for (int j = 0; j < nCols; j++) {
        	    			v[i][j] = listOfLines.get(i)[j];
        	    		}
        	    	}
        	    	results.setValuesTable(v);
        	    }
        	    

			} catch (FileNotFoundException e) { }
    	}
	}

	@Override
	protected List<Component> getPromptComponents(SElement fcmlElem, SElement algorithmElement, List<String> parameterNames)
	{
		// We need the plugin output folder and we want to add that to the algorithmElement so that later on, we can scan that folder for any existing .RData files.
		// Unfortunately, FJPluginHelper.getPluginOutputFolder(fcmlElem, this) seems to be returning null sometimes, 
		// i.e., this may be called before the plugin output folder is set. Therefore, this is a work around:
		Sample sample = FJPluginHelper.getSample(fcmlElem);
		if (sample != null) {
			Workspace ws = sample.getWorkspace();
			if (ws != null) {
				try {
					WSDocument wsd = ws.getDoc();
					String wsDir = wsd.getWorkspaceDirectory().getAbsolutePath();
					String wsName = wsd.getFilename();
					String outputFolder = wsDir + File.separator + wsName.substring(0, wsName.lastIndexOf('.')) + File.separator + this.getName();
					File pluginFolder = new File(outputFolder);
					if (pluginFolder.exists())
						algorithmElement.setAttribute(pluginFolderAttName, outputFolder);
				} catch (Exception e) {};
			}
		}
		List<Component> ret = super.getPromptComponents(fcmlElem, algorithmElement, parameterNames);
		refreshPlotChannelsComboOptions();
		if (fParameterNameList != null) fParameterNameList.addListSelectionListener(new ListSelectionListener() {
			@Override
			public void valueChanged(ListSelectionEvent e)
			{
				refreshPlotChannelsComboOptions();
			}
		});
		return ret;
	}
	
	@SuppressWarnings({"deprecation", "unchecked"})
	private void refreshPlotChannelsComboOptions()
	{
		if (fPlotChannelsCombo == null || fParameterNameList == null) return;
		String selectedPar = "";
		if (fPlotChannelsCombo.getSelectedItem() != null) 
			selectedPar = fPlotChannelsCombo.getSelectedItem().toString();
		fPlotChannelsCombo.removeAllItems();
		fPlotChannelsCombo.addItem(allChannelsText);
		fPlotChannelsCombo.setSelectedIndex(0);
		int i = 1;
		for (Object obj : fParameterNameList.getSelectedValues()) {
			String itemToAdd = obj.toString();
    		fPlotChannelsCombo.addItem(itemToAdd);
    		if (itemToAdd.compareTo(selectedPar) == 0)
    			fPlotChannelsCombo.setSelectedIndex(i);
    		i++;
		}
	}

	@SuppressWarnings("unchecked")
	@Override
	protected List<Component> getPromptComponents(SElement selement, List<String> list)
    {
		fsElement = selement;
		ArrayList<Component> componentList = new ArrayList<Component>();
        
        FJLabel fjLabel1 = new FJLabel(channelsLabelLine1);
        FJLabel fjLabel2 = new FJLabel(channelsLabelLine2);
        
        fParameterNameList = new FJList(new DefaultListModel<Object>());
        fParameterNameList.setSelectionMode(2);
        JScrollPane parListScrollPane = new JScrollPane(fParameterNameList);
        GuiFactory.setSizes(parListScrollPane, new Dimension(100, Math.min(200, list.size() * 20)));
        componentList.add(fjLabel1);
        componentList.add(fjLabel2);
        componentList.add(parListScrollPane);
        
        int nPars = list.size();
        String plotOptions[] = new String[nPars + 1];
        plotOptions[0] = allChannelsText;
        int i = 1;
        for (String p : list) 
    		plotOptions[i++] = p; 
        fPlotChannelsCombo = new FJComboBox(plotOptions);
        fPlotChannelsCombo = new FJComboBox();
        
        fApplyOnPrevCombo = new FJComboBox(new String[] { defaultApplyOnPrev });
        String pluginFolder = selement.getString(pluginFolderAttName);
        if (pluginFolder != null && !pluginFolder.isEmpty()) {
        	File myDir = new File(pluginFolder);
        	if (myDir.exists()) {
        		File[] existingRDataFiles = myDir.listFiles(new FilenameFilter() {
        		    public boolean accept(File dir, String name) {
        		        return name.endsWith(FlowSOM.RDataFileExtension);
        		    }
        		});
        		for (File rDataFile : existingRDataFiles) {
        			String rDataName = rDataFile.getName();
        			if (rDataName.contains(FlowSOM.RDataFileSuffix)) 
        				rDataName = rDataName.substring(0, rDataName.lastIndexOf(FlowSOM.RDataFileSuffix));
        			fApplyOnPrevCombo.addItem(rDataName);        			
        		}
        	}
        }
        
        // Default parameter values
        fnclus = defaultNumClust;
        fMetaclustersBackground = defaultMetaclustersBackground;
        fAddMFIsToTable = FlowSOM.defaultAddMFIsToTable;
        fndimx = defaultXDim;
        fndimy = defaultYDim;
        fScale = dafaultScale;
        fCreatePdf = dafaultCreatePDF;
        fFixedNodeSize = defaultFixedNodeSize;
        fNodeScale = defaultNodeScale;
        
        // If there are option set already (e.g., from the workspace), then
        // let's retrieve those and use them instead of defaults.
        Iterator<SElement> iterator = selement.getChildren("Option").iterator();
        int savedViewOptionIndex = 0;
        int savedPlotOptionIndex = 0;
        
        while(iterator.hasNext()) {
            SElement option = iterator.next();

            int savednClus = option.getInt(numClustOptionName, -1);
            if(savednClus > 0 && savednClus <= 900) fnclus = savednClus;
            
            int savedDimx = option.getInt(xDimOptionName, -1);
            if(savedDimx >= 3 && savedDimx <= 30) fndimx = savedDimx;

            int savedDimy = option.getInt(yDimOptionName, -1);
            if(savedDimy >= 3 && savedDimy <= 30) fndimy = savedDimy;
            
            double savedNodeScale = option.getDouble(nodeScaleOptionName, -1);
            if(savedNodeScale > 0 && savedNodeScale <= 1000.0) fNodeScale = savedNodeScale;

            String savedViewOption = option.getString(viewOptionName);
        	i = 0;
    		if (savedViewOption != null && !savedViewOption.isEmpty()) for (String s : viewComboOptions) {
    			if (s.equals(savedViewOption)) { 
    				savedViewOptionIndex = i;
    				continue;
    			}
    			i++;
    		}
    		
    		String savedPlotOption = option.getString(plotOptionName);
        	i = 0;
    		if (savedPlotOption != null && !savedPlotOption.isEmpty()) for (String s : plotOptions) {
    			if (s.equals(savedPlotOption)) { 
    				savedPlotOptionIndex = i;
    				continue;
    			}
    			i++;
    		}
    		
    		int nApplyOnPrevComboItemsCount = fApplyOnPrevCombo.getItemCount();
    		String savedApplyOnPrevOption = option.getString(applyOnPrevOptionName);
    		if (savedApplyOnPrevOption != null && savedApplyOnPrevOption.length() > 5) for (int j = 0; j < nApplyOnPrevComboItemsCount; j++) {
    			String itemValue = (String) fApplyOnPrevCombo.getItemAt(j);
    			if (savedApplyOnPrevOption.startsWith(itemValue)) 
    				fApplyOnPrevCombo.setSelectedIndex(j);
    		}
    		
    		String savedbgMetaClusters = option.getAttributeValue(bgMetaClustersOptionName);
    		if (savedbgMetaClusters != null && !savedbgMetaClusters.isEmpty())
    			fMetaclustersBackground = One.equals(savedbgMetaClusters) || True.equals(savedbgMetaClusters);

    		String savedAddMFIsToTable = option.getAttributeValue(addMFIsTableOptionName);
    		if (savedAddMFIsToTable != null && !savedAddMFIsToTable.isEmpty())
    			fAddMFIsToTable = One.equals(savedAddMFIsToTable) || True.equals(savedAddMFIsToTable);
    		
    		String savedScale = option.getAttributeValue(scaleOptionName);
    		if (savedScale != null && !savedScale.isEmpty())
    			fScale = One.equals(savedScale) || True.equals(savedScale);
    		
    		String savedCreatePdf = option.getAttributeValue(createPdfOptionName);
    		if (savedCreatePdf != null && !savedCreatePdf.isEmpty())
    			fCreatePdf = One.equals(savedCreatePdf) || True.equals(savedCreatePdf);
    		
    		String savedFixedNodeSize = option.getAttributeValue(fixedNodeSizeOptionName);
    		if (savedFixedNodeSize != null && !savedFixedNodeSize.isEmpty())
    			fFixedNodeSize = One.equals(savedFixedNodeSize) || True.equals(savedFixedNodeSize);
    		
        }
		if (savedViewOptionIndex >= 0 && savedViewOptionIndex < fViewCombo.getItemCount())
			fViewCombo.setSelectedIndex(savedViewOptionIndex);
		if (savedPlotOptionIndex >= 0 && savedPlotOptionIndex < fPlotChannelsCombo.getItemCount())
			fPlotChannelsCombo.setSelectedIndex(savedPlotOptionIndex);
		
		fApplyOnPrevCombo.addItemListener (new ItemListener () {
			@Override
			public void itemStateChanged(ItemEvent e)
			{
	    		refreshComponentsEnabled(fApplyOnPrevCombo);
			}
		});
		refreshComponentsEnabled(fApplyOnPrevCombo);

		FJLabel LabelApplyOnPrev = new FJLabel(applyOnPrevLabel);
        fApplyOnPrevCombo.setToolTipText("<html><p width=\"" + fixedToolTipWidth + "\">" + applyOnPrevTooltip + "</p></html>");
        GuiFactory.setSizes(fApplyOnPrevCombo, new Dimension(fixedComboWidth, fixedFieldHeigth));
        GuiFactory.setSizes(LabelApplyOnPrev, new Dimension(fixedLabelWidth, fixedLabelHeigth));
        HBox hboxApplyOnPrev = new HBox(new Component[] { LabelApplyOnPrev, fApplyOnPrevCombo });        
        componentList.add(hboxApplyOnPrev);
		componentList.add(new HBox(new Component[] { new FJLabel(orPerformFlowSOMLabel) }));
		
        FJLabel fjLabelNumClusters = new FJLabel(numClustLabel);
        fNumClustField = new RangedIntegerTextField(1, 90); // TODO: Ask Sofie, large values give errors, but the maximum is unclear (comes from cutree function, which somehow relates to input)
        fNumClustField.setInt(fnclus);
        fNumClustField.setToolTipText("<html><p width=\"" + fixedToolTipWidth + "\">" + numClustLabelTooltip + "</p></html>");
        GuiFactory.setSizes(fNumClustField, new Dimension(fixedFieldWidth, fixedFieldHeigth));
        GuiFactory.setSizes(fjLabelNumClusters, new Dimension(fixedLabelWidth, fixedLabelHeigth));
        HBox hboxNumClusters = new HBox(new Component[] { fjLabelNumClusters, fNumClustField });        
        componentList.add(hboxNumClusters);
        
        FJLabel fjLabelDimX = new FJLabel(dimXLabel);
        fDimXField = new RangedIntegerTextField(3, 30);
        fDimXField.setInt(fndimx);
        fDimXField.setToolTipText("<html><p width=\"" + fixedToolTipWidth + "\">" + dimXTooltip + "</p></html>");
        GuiFactory.setSizes(fDimXField, new Dimension(fixedFieldWidth, fixedFieldHeigth));
        GuiFactory.setSizes(fjLabelDimX, new Dimension(fixedLabelWidth, fixedLabelHeigth));

        fDimYField = new RangedIntegerTextField(3, 30);
        fDimYField.setInt(fndimy);
        fDimYField.setToolTipText("<html><p width=\"" + fixedToolTipWidth + "\">" + dimYTooltip + "</p></html>");
        GuiFactory.setSizes(fDimYField, new Dimension(fixedFieldWidth, fixedFieldHeigth));
        
        HBox hboxDimXY = new HBox(new Component[] { fjLabelDimX, fDimXField, /*fjLabelDimY,*/new FJLabel("x"), fDimYField });        
        componentList.add(hboxDimXY);
        
        fScaleOptionCheckbox = new FJCheckBox(scaleLabel);
        fScaleOptionCheckbox.setToolTipText("<html><p width=\"" + fixedToolTipWidth + "\">" + scaleTooltip + "</p></html>");
        fScaleOptionCheckbox.setSelected(fScale);
        componentList.add(new HBox(new Component[] { fScaleOptionCheckbox }));
        
        componentList.add(new HBox(new Component[] { new FJLabel(displayLabel) }));

        FJLabel LabelView = new FJLabel(viewLabel);
        fViewCombo.setToolTipText("<html><p width=\"" + fixedToolTipWidth + "\">" + viewTooltip + "</p></html>");
        GuiFactory.setSizes(fViewCombo, new Dimension(fixedComboWidth, fixedFieldHeigth));
        GuiFactory.setSizes(LabelView, new Dimension(fixedLabelWidth, fixedLabelHeigth));
        HBox hboxView = new HBox(new Component[] { LabelView, fViewCombo });        
        componentList.add(hboxView);
        
        FJLabel LabelPlot = new FJLabel(plotLabel);
        fPlotChannelsCombo.setToolTipText("<html><p width=\"" + fixedToolTipWidth + "\">" + plotTooltip + "</p></html>");
        GuiFactory.setSizes(fPlotChannelsCombo, new Dimension(fixedComboWidth, fixedFieldHeigth));
        GuiFactory.setSizes(LabelPlot, new Dimension(fixedLabelWidth, fixedLabelHeigth));
        HBox hboxPlot = new HBox(new Component[] { LabelPlot, fPlotChannelsCombo });        
        componentList.add(hboxPlot);
        
        FJLabel fjLabelNodeScale = new FJLabel(nodeScaleLabel);
        fNodeScaleField = new RangedDoubleTextField(0.01, 1000.0, new DoubleNumberFormatter());
        fNodeScaleField.setDouble(fNodeScale);
        fNodeScaleField.setToolTipText("<html><p width=\"" + fixedToolTipWidth + "\">" + nodeScaleTooltip + "</p></html>");
        GuiFactory.setSizes(fNodeScaleField, new Dimension(fixedFieldWidth, fixedFieldHeigth));
        GuiFactory.setSizes(fjLabelNodeScale, new Dimension(fixedLabelWidth, fixedLabelHeigth));
        HBox hboxNodeScale = new HBox(new Component[] { fjLabelNodeScale, fNodeScaleField, new FJLabel("%") });        
        componentList.add(hboxNodeScale);
        
        fMetaclustersBackgroundCheckbox = new FJCheckBox(backgroundLabel);
        fMetaclustersBackgroundCheckbox.setToolTipText("<html><p width=\"" + fixedToolTipWidth + "\">" + backgroundTooltip + "</p></html>");
        fMetaclustersBackgroundCheckbox.setSelected(fMetaclustersBackground);
        fFixedNodeSizeOptionCheckbox = new FJCheckBox(fixedNodeSizeLabel);
        fFixedNodeSizeOptionCheckbox.setToolTipText("<html><p width=\"" + fixedToolTipWidth + "\">" + fixedNodeSizeTooltip + "</p></html>");
        fFixedNodeSizeOptionCheckbox.setSelected(fFixedNodeSize);
        fCreatePdfOptionCheckbox = new FJCheckBox(createPdfLabel);
        fCreatePdfOptionCheckbox.setToolTipText("<html><p width=\"" + fixedToolTipWidth + "\">" + createPdfTooltip + "</p></html>");
        fCreatePdfOptionCheckbox.setSelected(fCreatePdf);
        HBox hboxCheckBoxes = new HBox(new Component[] { fMetaclustersBackgroundCheckbox, fFixedNodeSizeOptionCheckbox, fCreatePdfOptionCheckbox });
        componentList.add(hboxCheckBoxes);
        
        fAddMFIsTableCheckbox = new FJCheckBox(addMFIsTableLabel);
        fAddMFIsTableCheckbox.setToolTipText("<html><p width=\"" + fixedToolTipWidth + "\">" + addMFIsTableTooltip + "</p></html>");
        fShowOutputCheckBox = new JCheckBox(keepResultsLabel);
        fShowOutputCheckBox.setToolTipText("Keep a file that shows execution of the script");
        fShowOutputCheckBox.setToolTipText("<html><p width=\"" + fixedToolTipWidth + "\">" + keepResultsTooltip + "</p></html>");
        fShowOutputCheckBox.setSelected(fShowOutput);
        componentList.add(new HBox(new Component[] { fAddMFIsTableCheckbox, fShowOutputCheckBox }));
    	
		if (!fRoutFile.isEmpty())
			componentList.add(new HBox(Box.createHorizontalGlue(),  createShowOutputButton(), Box.createHorizontalGlue()));

		FJLabel hSpaceLabelCiting = new FJLabel("");
        GuiFactory.setSizes(hSpaceLabelCiting, new Dimension(fixedLabelWidth, hSpaceHeigth));
        componentList.add(hSpaceLabelCiting);
        
        componentList.add(new FJLabel(citingLabelLine1));
        componentList.add(new FJLabel(citingLabelLine2));
        componentList.add(new FJLabel(citingLabelLine3));
        componentList.add(new FJLabel(citingLabelLine4));
        
        return componentList;
    }
    
	protected void refreshComponentsEnabled(FJComboBox applyOnPrevCombo)
	{
		if (applyOnPrevCombo != null) {
			if (applyOnPrevCombo.getSelectedIndex() > 0) {
				if (fNumClustField != null) fNumClustField.setEnabled(false);
				if (fDimXField != null) fDimXField.setEnabled(false);
				if (fDimYField != null) fDimYField.setEnabled(false);
				if (fScaleOptionCheckbox != null) fScaleOptionCheckbox.setEnabled(false);

				// If we are selecting the application on existing map then let's select the same parameters
				// in the parameter selector, and let's do so based on the CSV file that has those.
				String csvWithParsName = applyOnPrevCombo.getSelectedItem().toString() + FlowSOM.CSVwithParsFileSuffix;
		        String pluginFolder = this.fsElement.getString(pluginFolderAttName);
		        if (pluginFolder != null && !pluginFolder.isEmpty() && fParameterNameList != null) {
		        	File csvWithParsFile = new File(pluginFolder + File.separator + csvWithParsName);
		        	if (csvWithParsFile.exists()) {
		        		ArrayList<String> parsList = new ArrayList<String>();
		        	    Scanner scan;
						try {
							scan = new Scanner(csvWithParsFile);
							if(scan.hasNextLine()) 
								scan.nextLine(); // Skip header (which is "x")
			        	    while (scan.hasNextLine()) {
			        	        String line = scan.nextLine();
			        	        line = line.replaceAll("^\"|\"$", "");
			        	        if (line != null && !line.isEmpty())
			        	        	parsList.add(line);
			        	    }
			        	    
			        	    List<Integer> selections = new ArrayList<Integer>();
			        	    @SuppressWarnings("rawtypes")
							DefaultListModel model = (DefaultListModel) fParameterNameList.getModel();
			        	    int nPars = model.getSize();
			        	    for (int i = 0; i < nPars; i++) {
			        	    	String pName = model.get(i).toString();
			        	    	if (pName != null && !pName.isEmpty()) {
			        	    		for (String nameFromCSV : parsList) {
			        	    			if (pName.startsWith(nameFromCSV)) {
			        	    				selections.add(i); 
			        	    				break;
			        	    			}
			        	    		}
			        	    	}
			        	    }
			        		int indices[] = new int[selections.size()];
			        		for (int i = 0; i < selections.size(); i++)
			        			indices[i] = selections.get(i);
			        		fParameterNameList.setSelectedIndices(indices);
			        		fParameterNameList.setEnabled(false);
						} catch (FileNotFoundException e) { }
		        	}
		        }
			} else {
				if (fNumClustField != null) fNumClustField.setEnabled(true);
				if (fDimXField != null) fDimXField.setEnabled(true);
				if (fDimYField != null) fDimYField.setEnabled(true);
				if (fScaleOptionCheckbox != null) fScaleOptionCheckbox.setEnabled(true);
				if (fParameterNameList != null) fParameterNameList.setEnabled(true);
			}
			refreshPlotChannelsComboOptions();
		}
	}

	@SuppressWarnings("deprecation")
	@Override
    protected void extractPromptOptions()
    {
    	fOptions = new HashMap<String, String>();
    	fParameterNames = new ArrayList<String>();

    	String selectedPlotChannel = allChannelsText;
    	if (fPlotChannelsCombo != null) selectedPlotChannel = fPlotChannelsCombo.getSelectedItem().toString();
    	int selectedPlotChannelparDescIndex = selectedPlotChannel.indexOf(" :: ");
		if(selectedPlotChannelparDescIndex > 0) selectedPlotChannel = selectedPlotChannel.substring(0, selectedPlotChannelparDescIndex);
    	boolean plotParameterIncluded = allChannelsText.equals(selectedPlotChannel);
    	boolean cellIdParameterIncluded = false;
    	for (Object obj : fParameterNameList.getSelectedValues()) 
    	{
    		String parName = (new StringBuilder()).append("").append(obj).toString();
    		// FlowJo's parameter names are often in the form of Name :: Description, we only want the Name part from that
    		int parDescIndex = parName.indexOf(" :: ");
    		if(parDescIndex > 0) parName = parName.substring(0, parDescIndex);
    		fParameterNames.add(parName);
    		if (parName.equals(selectedPlotChannel))
    			plotParameterIncluded = true;
    		if (parName.equals(FlowSOM.cellIdParName))
    			cellIdParameterIncluded = true;
    	}
    	// If plotParameter isn't among the list, we will switch to plotting all that's there
    	if(!plotParameterIncluded) 
    		selectedPlotChannel = allChannelsText;
    	// TODO
    	// We really need the Time parameter, so we select it even if the user doesn't.
    	if (isSeqGeq() && !cellIdParameterIncluded)
			fParameterNames.add(FlowSOM.cellIdParName);
    	
    	// Save all the FlowSOM specific options
    	fOptions.put(numClustOptionName, Integer.toString(fNumClustField.getInt()));
        fOptions.put(bgMetaClustersOptionName, fMetaclustersBackgroundCheckbox.isSelected() ? One : Zero);
        fOptions.put(addMFIsTableOptionName, fAddMFIsTableCheckbox.isSelected() ? One : Zero);
        fOptions.put(xDimOptionName, Integer.toString(fDimXField.getInt()));
        fOptions.put(yDimOptionName, Integer.toString(fDimYField.getInt()));
        fOptions.put(scaleOptionName, fScaleOptionCheckbox.isSelected() ? One : Zero);
        fOptions.put(createPdfOptionName, fCreatePdfOptionCheckbox.isSelected() ? One : Zero);
        fOptions.put(fixedNodeSizeOptionName, fFixedNodeSizeOptionCheckbox.isSelected() ? One : Zero);
        fOptions.put(nodeScaleOptionName, Double.toString(fNodeScaleField.getDouble()));
    	fOptions.put(plotOptionName, selectedPlotChannel);
    	if (fApplyOnPrevCombo.getSelectedIndex() <= 0) {
    		fOptions.put(applyOnPrevOptionName, fApplyOnPrevCombo.getSelectedItem().toString());
    	} else {
    		fOptions.put(applyOnPrevOptionName, fApplyOnPrevCombo.getSelectedItem().toString() + FlowSOM.RDataFileSuffix);
    	}
    	
    	
    	if (fViewCombo.getSelectedIndex() < viewComboOptions.length)
    		fOptions.put(viewOptionName, viewComboOptions[fViewCombo.getSelectedIndex()]);
    	else 
    		fOptions.put(viewOptionName, viewComboOptions[0]);
    	    	
    	fShowOutput = fShowOutputCheckBox.isSelected();
    }

	// Is this SeqGeq based on the application name?
	// This is used to figure out whether we should be including CellIds in the CSV files
	public static boolean isSeqGeq()
	{
        if (FJApplication.getInstance() != null && 
            FJApplication.getInstance().getAppName() != null && 
            FJApplication.getInstance().getAppName().toLowerCase().contains("seqgeq"))
            	return true;
		return false;
	}

}
