////////////////////////////////////////////////////////////////////////////////
// Copyright (c) 2017 Josef Spidlen, Ph.D., FlowJo, LLC
//
// License
// The software is distributed under the terms of the 
// Artistic License 2.0
// http://www.r-project.org/Licenses/Artistic-2.0
// 
// Disclaimer
// This software and documentation come with no warranties of any kind.
// This software is provided "as is" and any express or implied 
// warranties, including, but not limited to, the implied warranties of
// merchantability and fitness for a particular purpose are disclaimed.
// In no event shall the  copyright holder be liable for any direct, 
// indirect, incidental, special, exemplary, or consequential damages
// (including but not limited to, procurement of substitute goods or 
// services; loss of use, data or profits; or business interruption)
// however caused and on any theory of liability, whether in contract,
// strict liability, or tort arising in any way out of the use of this 
// software.    
//////////////////////////////////////////////////////////////////////////////

package com.flowjo.plugin.FlowSOM;

import java.io.BufferedReader;
import java.io.File;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.io.StringWriter;
import java.util.List;
import java.util.Map;

import com.treestar.lib.file.FileUtil;
import com.treestar.flowjo.engine.EngineManager;
import com.treestar.flowjo.engine.utility.RFlowCalculator;

import com.flowjo.plugin.FlowSOM.utils.FilenameUtils;

public class FlowSOMRFlowCalc extends RFlowCalculator {
	
	// The path to the flowSOM R script template from within the jar file
	private final static String flowSOMTemplatePath = "r/RScript.flowSOM.Template.R";
    public static final String TRUE = "TRUE";
    public static final String FALSE = "FALSE";

	
	public File runFlowSOM(File sampleFile, String sampleName, List<String> parameterNames, Map<String, String> options, String outputFolderPath, boolean useExistingFiles)
    {
		sampleName = sampleName.replaceAll(".ExtNode", "").replaceAll(".fcs", "").replaceAll(".LMD", "").trim();
        File outputFolder = new File(outputFolderPath);
        StringWriter scriptWriter = new StringWriter();
        File flowSOMScript = createFlowSOMscript(sampleFile, sampleName, parameterNames, options, outputFolder, scriptWriter);
        if(flowSOMScript == null) return null;
        if(useExistingFiles && flowSOMScript.exists()) return flowSOMScript;
        
        String scriptFileName = (new StringBuilder()).append("RScript.flowSOM.").append(System.currentTimeMillis()).append(".R").toString().replaceAll(" ", "_");
        try
        {
            fScriptFile = new File(outputFolderPath, scriptFileName);
            FileUtil.write(fScriptFile, scriptWriter.toString());
            executeRBatch(scriptFileName);
        }
        catch(Exception e) 
        {
            e.printStackTrace();
        }
        return flowSOMScript;
    }

	protected File createFlowSOMscript(File sampleFile, String sampleName, List<String> parameterNames, Map<String, String> options, File outputFolder, StringWriter scriptWriter)
    {
        InputStream scriptStream = FlowSOMRFlowCalc.class.getResourceAsStream(flowSOMTemplatePath);
        
        String outFileName = (new StringBuilder()).append(FilenameUtils.fixFileNamePart(sampleName)).append(".flowSOM").append(".csv").toString();
        if(outputFolder == null) outputFolder = sampleFile.getParentFile();
        
        File outFile = new File(outputFolder, outFileName);
        outFileName = outFile.getAbsolutePath();
        String dataFilePath = sampleFile.getAbsolutePath();

        if(EngineManager.isWindows()) outFileName = outFileName.replaceAll("\\\\", "/");
        if(EngineManager.isWindows()) dataFilePath = dataFilePath.replaceAll("\\\\", "/");

        String sNumClus = options.get(FlowSOM.numClustOptionName);
        String sViewOption = options.get(FlowSOM.viewOptionName);
        String sPlotOption = options.get(FlowSOM.plotOptionName);
        String sBgMetaClusters = options.get(FlowSOM.bgMetaClustersOptionName);
        String sParScale = options.get(FlowSOM.scaleOptionName);
        String sParCreatePdf = options.get(FlowSOM.createPdfOptionName);
        String sParFixedNodeSize = options.get(FlowSOM.fixedNodeSizeOptionName);
        String sPlotAll = FlowSOM.allChannelsText.equals(sPlotOption) ? TRUE : FALSE;
        String sParXDim = options.get(FlowSOM.xDimOptionName);
        String sParYDim = options.get(FlowSOM.yDimOptionName);
        String sParNodeScale = options.get(FlowSOM.nodeScaleOptionName);
        String sParApplyOnPrev = options.get(FlowSOM.applyOnPrevOptionName);        
        String sAddCellIdToResults = FlowSOM.isSeqGeq() ? TRUE : FALSE;
                
        if (sBgMetaClusters == null || sBgMetaClusters.isEmpty() || FlowSOM.One.equals(sBgMetaClusters) || FlowSOM.True.equals(sBgMetaClusters))
        	sBgMetaClusters = TRUE; // TRUE is the default
        else
        	sBgMetaClusters = FALSE;
        
        if (sParScale == null || sParScale.isEmpty() || FlowSOM.Zero.equals(sParScale) || FlowSOM.False.equals(sParScale))
        	sParScale = FALSE; // FALSE is the default
        else
        	sParScale = TRUE;
        
        if (sParCreatePdf == null || sParCreatePdf.isEmpty() || FlowSOM.One.equals(sParCreatePdf) || FlowSOM.True.equals(sParCreatePdf))
        	sParCreatePdf = TRUE; // TRUE is the default
        else
        	sParCreatePdf = FALSE;
        
        if (sParFixedNodeSize == null || sParFixedNodeSize.isEmpty() || FlowSOM.Zero.equals(sParFixedNodeSize) || FlowSOM.False.equals(sParFixedNodeSize))
        	sParFixedNodeSize = FALSE; // FALSE is the default
        else
        	sParFixedNodeSize = TRUE;
        
        if (sParApplyOnPrev == null || sParApplyOnPrev.isEmpty())
        	sParApplyOnPrev = FlowSOM.defaultApplyOnPrev;
        if (sParApplyOnPrev.length() > 5) // i.e., not "None"
        	sParApplyOnPrev = outputFolder + File.separator + sParApplyOnPrev;
        if(EngineManager.isWindows()) sParApplyOnPrev = sParApplyOnPrev.replaceAll("\\\\", "/");

        try {
        	if ((Integer.parseInt(sParXDim) < 3) || (Integer.parseInt(sParXDim) > 30))
        		sParXDim = Integer.toString(FlowSOM.defaultXDim);
        } catch (Exception e) {
        	sParXDim = Integer.toString(FlowSOM.defaultXDim);
        }
        
        try {
        	if ((Integer.parseInt(sParYDim) < 3) || (Integer.parseInt(sParYDim) > 30))
        		sParYDim = Integer.toString(FlowSOM.defaultYDim);
        } catch (Exception e) {
        	sParYDim = Integer.toString(FlowSOM.defaultYDim);
        }
        
        int xDim = FlowSOM.defaultXDim;
        int yDim = FlowSOM.defaultYDim;
        try {
        	xDim = Integer.parseInt(sParXDim);
        	yDim = Integer.parseInt(sParYDim);
        } catch (Exception e) {}
        
        try {
        	if ((Integer.parseInt(sNumClus) <= 0))
				sNumClus = Integer.toString(FlowSOM.defaultNumClust);
        } catch (Exception e) {
        	sNumClus = Integer.toString(FlowSOM.defaultNumClust);
        }        
        
        try {
        	// FlowSOM calls cutree and it appears the cutree enforces a maximum of floor(0.9 * xDim * yDim)
        	if (Integer.parseInt(sNumClus) > (int)Math.floor(0.9 * xDim * yDim))
				sNumClus = Integer.toString((int)Math.floor(0.9 * xDim * yDim));
        } catch (Exception e) {
        	sNumClus = Integer.toString(FlowSOM.defaultNumClust);
        }
        
        try {
        	if ((Double.parseDouble(sParNodeScale) <= 0 || Double.parseDouble(sParNodeScale) > 1000))
        		sParNodeScale = Double.toString(FlowSOM.defaultNodeScale);
        } catch (Exception e) {
        	sParNodeScale = Double.toString(FlowSOM.defaultNodeScale);
        }

        BufferedReader rTemplateReader = null;
        try {
			rTemplateReader = new BufferedReader(new InputStreamReader(scriptStream));
		} catch (Exception e) {
			e.printStackTrace();
			return null;
		}
        
        String scriptLine;
        try {
			while((scriptLine = rTemplateReader.readLine()) != null) 
			{
			    scriptLine = scriptLine.replaceAll("FJ_DATA_FILE_PATH", dataFilePath);
			    scriptLine = scriptLine.replaceAll("FJ_CSV_OUPUT_FILE", outFileName);
			    scriptLine = scriptLine.replaceAll("FJ_PAR_NUM_CLUST", sNumClus);
			    scriptLine = scriptLine.replaceAll("FJ_BG_META_CLUSTERS", sBgMetaClusters);
			    scriptLine = scriptLine.replaceAll("FJ_PLOT_OPTION", sPlotOption);
			    scriptLine = scriptLine.replaceAll("FJ_VIEW_OPTION", sViewOption);
			    scriptLine = scriptLine.replaceAll("FJ_PLOT_ALL", sPlotAll);
			    scriptLine = scriptLine.replaceAll("FJ_PAR_SCALE", sParScale);
			    scriptLine = scriptLine.replaceAll("FJ_PAR_CREATE_PDF", sParCreatePdf);
			    scriptLine = scriptLine.replaceAll("FJ_PAR_XDIM", sParXDim);
			    scriptLine = scriptLine.replaceAll("FJ_PAR_YDIM", sParYDim);
			    scriptLine = scriptLine.replaceAll("FJ_PAR_FIXED_NODE_SIZE", sParFixedNodeSize);
			    scriptLine = scriptLine.replaceAll("FJ_PAR_NODE_SCALE", sParNodeScale);
			    scriptLine = scriptLine.replaceAll("FJ_PAR_APPLY_ON_PREV", sParApplyOnPrev);
			    scriptLine = scriptLine.replaceAll("FJ_PAR_ADD_CELLIDS_TO_RESULT", sAddCellIdToResults);

			    if(scriptLine.contains("FJ_PARAMS_LIST")) {
			    	String parListStr = "";
			    	for (String parName : parameterNames)
			    	{
			    		// We don't want the TIME parameter to be in the parameter list given to the R script
			    		if(parName.compareToIgnoreCase("TIME") != 0) {
			    			if(!parListStr.isEmpty()) parListStr = (new StringBuilder()).append(parListStr).append(",").toString();
			    			parListStr = (new StringBuilder()).append(parListStr).append("\"").append(parName).append("\"").toString();
			    		}
			    	}
			    	scriptLine = scriptLine.replaceAll("FJ_PARAMS_LIST", parListStr);
			    }
			    scriptWriter.append(scriptLine).append('\n');
			}
		} catch (Exception e) {
			e.printStackTrace();
		}
        
        if(rTemplateReader != null) {
        	try { rTemplateReader.close(); }
        	catch (Exception e) { e.printStackTrace(); }
        }
        
        return outFile;
    }

}
